state_dict = {
    'Johor': [1, 21, 22, 23, 24],
    'Kedah': [2, 25, 26, 27],
    'Kelantan': [3, 28, 29],
    'Malacca': [4, 30],
    'Negeri Sembilan': [5, 31, 59],
    'Pahang': [6, 32, 33],
    'Penang': [7, 34, 35],
    'Perak': [8, 36, 37, 38, 39],
    'Perlis': [9, 40],
    'Selangor': [10, 41, 42, 43, 44],
    'Terengganu': [11, 45, 46],
    'Sabah': [12, 47, 48, 49],
    'Sarawak': [13, 50, 51, 52 ,53],
    'Federal Territory of Kuala Lumpur': [14, 54, 55, 56, 57],
    'Federal Territory of Labuan': [15, 58],
    'Federal Territory of Putrajaya': [16],
}

abbreviation_dict = {
    'Johor': 'JHR',
    'Kedah': 'KDH',
    'Kelantan': 'KTN',
    'Malacca': 'MLK',
    'Negeri Sembilan': 'NSN',
    'Pahang': 'PHG',
    'Penang': 'PNG',
    'Perak': 'PRK',
    'Perlis': 'PLS',
    'Sabah': 'SBH',
    'Sarawak': 'SWK',
    'Selangor': 'SGR',
    'Terengganu': 'TGR',
    'Kuala Lumpur': 'KUL',
    'Labuan': 'LBN',
    'Putrajaya': 'PJY'
}



from .utils import is_mykad_valid, get_birthplace
from datetime import datetime


class MyKad:
    """The base MyKad class.

    :param mykad_num: The MyKad number. This can contain numbers and '-'
    :type mykad_num: str, int
    """
    def __init__(self, mykad_num):
        if (is_mykad_valid(mykad_num)):
            self.mykad_num = mykad_num.replace('-', '')
        else:
            raise ValueError(f'MyKad number {mykad_num} is not valid')

    def __str__(self):
        return f'{self.get_formatted()}'

    @property
    def birthdate_yymmdd(self):
        return self.mykad_num[0:6]

    @property
    def birthplace_code(self):
        return self.mykad_num[6:8]

    @property
    def special_nrd_num(self):
        return self.mykad_num[8:11]

    @property
    def gender_code(self):
        return self.mykad_num[11]

    @property
    def birthdate(self):
        """Returns the datetime instance corresponding to this
        MyKad holder.

        :return: The datetime instance
        :rtype: datetime
        """
        birthdate = datetime.strptime(self.birthdate_yymmdd, "%y%m%d")

        # Make sure dates past current time are 19xx and not 20xx.
        # This is mostly due to the limitations of ISO 8601:2000
        # where a date '570620' in YYMMDD format could be initialised
        # as 20th June 2057 rather than 20th June 1957.
        # I don't see any workaround to this unless the Malaysian
        # gov switches to a YYYYMMDD format.
        if birthdate.year > datetime.now().year:
            # Reconstruct datetime instance with a different year (i.e. 19xx instead of 20xx)
            birthdate = datetime(year=int('19' + str(birthdate.year)[2:4]), month=birthdate.month, day=birthdate.day)

        return birthdate

    @property
    def birth_year(self):
        return self.birthdate.year

    @property
    def birth_month(self):
        return self.birthdate.month

    @property
    def day_of_birth(self):
        return self.birthdate.day

    @property
    def birthplace(self):
        return get_birthplace(self.birthplace_code)

    @property
    def gender(self):
        if self.is_male():
            return "Male"
        else:
            return "Female"

    def get_unformatted(self):
        """Returns the unformatted MyKad string (i.e. just numbers, without '-')

        :return: The unformatted MyKad number
        :rtype: str
        """
        return self.mykad_num

    def get_formatted(self):
        """Returns the formatted MyKad string (with '-')
                :return: The formatted MyKad number
        :rtype: str
        """
        return f'{self.birthdate.strftime("%y%m%d")}-{self.birthplace_code}-{self.special_nrd_num}{self.gender_code}'

    def is_male(self):
        """Checks if the MyKad holder is a male.

        :return True if male, False otherwise
        :rtype bool
        """
        return int(self.gender_code) % 2 != 0

    def is_female(self):
        """Checks if the MyKad holder is a female.

        :return True if female, False otherwise
        :rtype bool
        """
        return int(self.gender_code) % 2 == 0



from .constants import abbreviation_dict, state_dict


def is_mykad_valid(mykad_num):
    """Checks if a MyKad number is valid.

    :return: True if MyKad is valid, False otherwise
    :rtype: bool
    """

    # Since this is meant to be a generic utility function, we accept both str and int
    if type(mykad_num) != str and type(mykad_num) != int:
        return False

    mykad_num = str(mykad_num)

    if '-' in mykad_num:
        mykad_num = mykad_num.replace('-', '')

    # MyKad should be correct length
    if len(mykad_num) != 12:
        return False

    # MyKad should only contain decimals
    if not mykad_num.isdecimal():
        return False

    # Check if PB (place of birth) code is valid
    try:
        get_birthplace(mykad_num[6:8])
    except ValueError:
        return False

    return True

def get_birthplace(birthplace_code):
    """Returns the birthplace of the MyKad holder.

    :return The birthplace code of the MyKad holder (i.e. 'BP' in YYMMDD-BP-###G)
    :rtype str, int
    """
    if (int(birthplace_code) == 0 or
        int(birthplace_code) == 17 or
        int(birthplace_code) == 18 or
        int(birthplace_code) == 19 or
        int(birthplace_code) == 20 or
        int(birthplace_code) == 69 or
        int(birthplace_code) == 70 or
        int(birthplace_code) == 73 or
        int(birthplace_code) == 80 or
        int(birthplace_code) == 81 or
        int(birthplace_code) == 94 or
        int(birthplace_code) == 95 or
        int(birthplace_code) == 96 or
        int(birthplace_code) == 97):
        raise ValueError(f'code {birthplace_code} is an invalid birthplace code')

    for key, val in state_dict.items():
        if int(birthplace_code) in val:
            return key

    return 'Outside Malaysia'

def get_state_abbreviation(state):
    """Gets the state abbreviation.

    :param state: The name of the state. This can be lower or upper-case
    :type state: str

    :return: State abbreviation (i.e. SGR, KUL, etc.)
    :rtype: str
    """
    for key, val in abbreviation_dict.items():
        # Make it lowercase to better generalise it
        if state.lower() == key.lower():
            return val

    raise ValueError(f'unknown state {state}')
